#Copyright 2001 by R. David Murray and FCGNetworks (http://www.fcgnetworks.net)
#Governed by the Artistic License
#  http://www.opensource.org/licenses/artistic-license.html

#Version 0.0.1

"""
'readline' an object on demand as a pseudo-sequence analogous to xrange

This wrapper adds to the methods available on a regular file object,
providing convenient ways to walk through the file processing the lines.

An xfile behaves as if it were an imutable sequence consisting
of lines selected from the file, so you can do, for example:

for line in readlinexrange(stdin):
	[suite]

If a line beyond the end of the file is accessed, an IndexError is
raised. 

Only seekable files open for reading or appending should be used
as xfiles; the results of accessing an xfile after an internal
write has been done are undefined.  For this reason, xfile does
*not* implement write operations.  (Note: a non-seekable file may
be used as long as successive accesses to the xfile use only
increasing indexes.)

The value returned by sequence access is equivalent to line, below. 

Creation:

xfile(fileobject,selector=None,parser=None)

  fileobject is any object that implements a file-like interface.
    It may also be a string, in which case the string is passed to 'open'
    to create the necessary file object.
  selector is a one argument boolean function.  If it returns false, the
    line is *not* included in the sequence constructed from the file.
    The default selector always returns true.
  parser is a one argument function that will be applied to each line
    before it is returned.  The default parser simply removes the
    newline from the end of the line.  The parser is *not* applied
    to lines returned by readline or raw. 


Methods:

raw()			Current line of file, unmodifed (has trailing newline).
setParser(parser)	Change parser.
lineno()		Line number of current line within source file.
line()			Current line after processing by parser.
next()			Line line, but file pointer is incramented first.
readline()		Like next, but returns raw.
nextnonblank()		Return next line() containing non-whitespace.
nextblank()		Return next line() consisting of only whitespace.
seek(i)			Equivalent to using a non-negative i as an index.
tell()			Returns sequence index of line after the one
                          last referenced.

Implementation notes:

Reading and parsing is delayed as long as possible (lazy evaluation).
Operations such as taking the len of the xfile or using a negative
index will cause the entire file to be read.  Parsing of lines is
done on demand.  Only seek indexes are kept in memory, so accessing
an xfile in random access fashion uses a minimal amount of memory
for even a fairly large selected line set.

"""

from string import strip

def stripnl(line):
  if not line: return line
  if line[-1]=='\n': return line[:-1]
  return line

class xfile:

  def __init__(self,file,selector=lambda x: 1,parser=stripnl):
    if type(file)==type(''): file = open(file)
    self._sourcefile = file
    self._filename = getattr(file,'name','<unknown>')
    self._selector = selector
    if parser is None: parser = self.stripnl
    self.setParser(parser)
    self._curline = -1
    self._rawline = ''
    self._lineno = -1
    self._linetable = []
    self._allread = 0

  def seek(self,i):
     if i==self._curline: return
     if i<0: raise IndexError, "Attempt to seek before start of xfile"
     if i<len(self._linetable):
       pos, self._lineno = self._linetable[i]
       if pos is None: raise IOError("Attempt to seek on non-seekable file")
       self._sourcefile.seek(pos)
       self._rawline = self._sourcefile.readline()
       self._curline = i
       return
     if self._allread: raise IndexError, "end of file reached"
     if self._lineno>=0 and self._lineno<>self._linetable[-1][1]:
       self.seek(len(self._linetable)-1)
     while i>=len(self._linetable):
       if hasattr(self._sourcefile,'tell'):
         try: pos = self._sourcefile.tell()
         except IOError: pos = None
       else: pos = None
       raw = self._sourcefile.readline()
       if raw == '':
         self._allread = 1
         raise IndexError, "end of file reached"
       self._rawline = raw
       self._lineno = self._lineno + 1
       if self._selector(self._rawline):
         self._linetable.append((pos,self._lineno))
     self._curline = i
     return

  def _readall(self):
    if self._allread: return
    last = len(self._linetable)-1
    if last>=0: self.seek(last)
    try:
      while 1: self.readline()
    except IndexError: pass

  def __repr__(self):
    return "<xfile('%s',selector=%s,parser=%s);curline(%s)=%s,maxread=%s>" % \
      (self._filename,self._selector,self._parser,self._curline,
       repr(self._rawline),len(self._linetable))

  def __str__(self):
    return "xfile('%s')" % self._filename

  def __getitem__(self,i):
    if i<0:
      self._readall()
      i = len(self._linetable)+i
      if i<0: raise IndexError("Index out of range")
    self.seek(i)
    return self.line()

  def __getslice__(self,i,j):
    if i<0 or j<0:
      self._readall()
      if i<0: i = len(self._linetable)+i
      if j<0: j = len(self._linetable)+j
    res = []
    for x in xrange(i,j):
      try: res.append(self[x])
      except IndexError: return res
    return res
    
  def __contains__(self, item):
    for line in self:
      if line==item: return 1
    return 0

  def count(self,item):
    c = 0
    for line in self:
      if line==item: c = c + 1
    return c

  def index(self,item):
    i = 0
    for line in self:
      if line==item: return i
      i = i + 1
    raise ValueError("%s not in file" % repr(item))

  def tell(self): return self._curline+1
  def _readline(self): self.seek(self._curline+1)
  def lineno(self): return self._lineno+1
  def raw(self): return self._rawline
  def raw_nonl(self): return stripnl(self._rawline)
  def setParser(self,parser): self._parser = parser
  def line(self): return self._parser(self.raw())
  def __add__(self,other): return self[:]+other
  def __radd__(self,other): return other+self[:]
  def __mul__(self,n): return self[:]*n

  def __len__(self): 
    self._readall()
    return len(self._linetable)

  def next(self):
    self._readline()
    return self.line()

  def readline(self):
    self._readline()
    return self.raw()

  def nextnonblank(self):
    while 1:
      self._readline()
      if strip(self.raw_nonl()): return self.line()

  def nextblank(self):
    while 1:
      self._readline()
      if not strip(self.raw_nonl()): return self.line()
